from .common import wer, per, cer
from .der import der, der_for_dynamic_superb
from .signal import pesq, stoi, sisdri
from .translation import sacre_bleu
from .slot_filling import slot_type_f1, slot_value_cer


"""Tests for metrics."""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import itertools
import os

import numpy as np
from numpy.testing import (assert_almost_equal, assert_equal,
                           assert_raises_regex)

from scorelib.metrics import (bcubed, conditional_entropy, contingency_matrix,
                              der, jer, goodman_kruskal_tau,
                              mutual_information)
from scorelib.rttm import load_rttm
from scorelib.score import turns_to_frames


TEST_DIR = os.path.dirname(os.path.abspath(__file__))


def make_labels(n=1000, n_classes=3, one_hot=False, seed=99999):
    """Return some cluster labelings for a toy dataset."""
    rstate = np.random.RandomState(seed)
    def _one_hot(x):
        classes, x = np.unique(x, return_inverse=True)
        return np.column_stack([x == c for c in classes])
    ref_labels = rstate.randint(n_classes, size=n)
    sys_labels = rstate.randint(n_classes, size=n)
    if one_hot:
        ref_labels = _one_hot(ref_labels)
        sys_labels = _one_hot(sys_labels)
    return ref_labels, sys_labels


def test_contingency_matrix():
    # Test exceptions.
    with assert_raises_regex(
            ValueError, 'ref_labels and sys_labels should either both be 1D'):
        contingency_matrix(np.zeros(5), np.zeros((5, 2)))
    with assert_raises_regex(
            ValueError, 'ref_labels and sys_labels must have same size'):
        contingency_matrix(np.arange(5), np.arange(6))

    # Test 1-D inputs.
    X, Y = make_labels()
    cm = contingency_matrix(X, Y)
    cm_expected = np.array(
        [[106, 114, 117],
         [110, 130, 105],
         [92, 118, 108]])
    assert_equal(cm, cm_expected)

    # Test 2-D inputs.
    X, Y = make_labels(one_hot=True)
    cm = contingency_matrix(X, Y)
    assert_equal(cm, cm_expected)


def test_bcubed():
    x, y = make_labels()

    # Test from inputs.
    p, r, f1 = bcubed(x, y)
    assert_almost_equal(p, 0.3345, 3)
    assert_almost_equal(r, 0.3356, 3)
    assert_almost_equal(f1, 0.3351, 3)

    # Test from CM.
    cm = contingency_matrix(x, y)
    p, r, f1 = bcubed(None, None, cm)
    assert_almost_equal(p, 0.3345, 3)
    assert_almost_equal(r, 0.3356, 3)
    assert_almost_equal(f1, 0.3351, 3)


def test_goodman_kruskal_tau():
    x, y = make_labels()

    # Test from inputs.
    tau_rs, tau_sr = goodman_kruskal_tau(x, y)
    assert_almost_equal(tau_rs, 0.001223, 5)
    assert_almost_equal(tau_sr, 0.001223, 5)

    # Test from CM.
    cm = contingency_matrix(x, y)
    tau_rs, tau_sr = goodman_kruskal_tau(None, None, cm)
    assert_almost_equal(tau_rs, 0.001223, 5)
    assert_almost_equal(tau_sr, 0.001223, 5)


def test_conditional_entropy():
    x, y = make_labels()

    # Test from inputs.
    ce = conditional_entropy(x, y)
    assert_almost_equal(ce, 1.5824, 3)

    # Test from CM.
    cm = contingency_matrix(x, y)
    ce = conditional_entropy(None, None, cm)
    assert_almost_equal(ce, 1.5824, 3)


def test_mutual_information():
    x, y = make_labels()

    # Test from inputs.
    mi, nmi = mutual_information(x, y)
    assert_almost_equal(mi, 0.001767, 5)
    assert_almost_equal(nmi, 0.001116, 5)

    # Test from CM.
    cm = contingency_matrix(x, y)
    mi, nmi = mutual_information(None, None, cm)
    assert_almost_equal(mi, 0.001767, 5)
    assert_almost_equal(nmi, 0.001116, 5)


def test_der():
    ref_turns, _, _ = load_rttm(
        os.path.join(TEST_DIR, 'ref.rttm'))
    sys_turns, _, _ = load_rttm(
        os.path.join(TEST_DIR, 'sys.rttm'))
    expected_der = 26.3931
    file_to_der, global_der = der(ref_turns, sys_turns)
    assert_almost_equal(file_to_der['FILE1'], expected_der, 3)
    assert_almost_equal(global_der, expected_der, 3)


def test_jer():
    # Check input validation.
    with assert_raises_regex(
            ValueError, 'All passed dicts must have same'):
        jer(dict(), dict(F1=np.zeros(1)), dict())
        jer(dict(F1=np.zeros(1)), dict(), dict())
        jer(dict(), dict(), dict(F1=np.zeros(1)))

    # Edge case: no reference speech.
    ref_durs = np.array([], dtype='int64')
    sys_durs = np.array([5, 5], dtype='int64')
    cm = np.zeros((0, 2), dtype='int64')
    file_to_jer, global_jer = jer(
        dict(F=ref_durs), dict(F=sys_durs), dict(F=cm))
    assert file_to_jer['F'] == 100.
    assert global_jer == 100.

    # Edge case: no system speech.
    ref_durs = np.array([5, 5], dtype='int64')
    sys_durs = np.array([], dtype='int64')
    cm = np.zeros((2, 0), dtype='int64')
    file_to_jer, global_jer = jer(
        dict(F=ref_durs), dict(F=sys_durs), dict(F=cm))
    assert file_to_jer['F'] == 100.
    assert global_jer == 100.

    # Edge case: no reference OR system speech.
    ref_durs = np.array([], dtype='int64')
    sys_durs = np.array([], dtype='int64')
    cm = np.zeros((0, 0), dtype='int64')
    file_to_jer, global_jer = jer(
        dict(F=ref_durs), dict(F=sys_durs), dict(F=cm))
    assert file_to_jer['F'] == 0.
    assert global_jer == 0.

    # Real data.
    ref_turns, _, _ = load_rttm(
        os.path.join(TEST_DIR, 'ref.rttm'))
    sys_turns, _, _ = load_rttm(
        os.path.join(TEST_DIR, 'sys.rttm'))
    dur = 1 + max(turn.offset
                  for turn in itertools.chain(ref_turns, sys_turns))
    ref_labels = turns_to_frames(ref_turns, [(0, dur)], step=0.01)
    sys_labels = turns_to_frames(sys_turns, [(0, dur)], step=0.01)
    cm = contingency_matrix(ref_labels, sys_labels)
    ref_durs = ref_labels.sum(axis=0)
    sys_durs = sys_labels.sum(axis=0)
    file_to_jer, global_jer = jer(
        dict(FILE1=ref_durs), dict(FILE1=sys_durs), dict(FILE1=cm))
    assert_almost_equal(file_to_jer['FILE1'], 33.24631, 3)
    assert_almost_equal(global_jer, 33.24631, 3)



from speech_metrics import slot_type_f1, slot_value_cer

def test_slot_filling():
    hyps = [
        "music_event: concert, music_location: New York, music_time: Tomorrow",
        "music_event: concert, music_location: New York, music_time: Tomorrow",
        "music_event: concert, music_location: New York, music_time: Tomorrow",
    ]
    refs = [
        "music_event: concert, music_location: new york, music_time: tomorrow",
        "music_event: concert, music_location: New York, music_time: Tomorrow",
        "music_event: Concert, music_location: New York",
    ]
    print(slot_type_f1(hyps, refs))
    print(slot_value_cer(hyps, refs))



"""Functions for scoring paired system/reference RTTM files."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from collections import defaultdict, namedtuple

import numpy as np
from scipy.linalg import block_diag

from . import metrics
from .six import iteritems, itervalues
from .utils import groupby

__all__ = ['flatten_labels', 'score', 'turns_to_frames', 'Scores']


def turns_to_frames(turns, score_regions, step=0.010):
    """Return frame-level labels corresponding to diarization.

    Parameters
    ----------
    turns : list of Turn
        Speaker turns. Should all be from single file.

    score_regions : list of tuple
        Scoring regions from UEM.

    step : float, optional
        Frame step size in seconds.
        (Default: 0.01)

    Returns
    -------
    labels : ndarray, (n_frames, n_speakers)
        Frame-level labels. The ``i,j``-th entry of this array is 1 if speaker
        ``j``-th speaker was present at frame ``i`` and 0 otherwise. If no
        speaker turns were passed, the second dimension will be 0.
    """
    file_ids = {turn.file_id for turn in turns}
    if len(file_ids) > 1:
        raise ValueError('Turns should be from a single file.')

    # Create matrix whose i,j-th entry is True IFF the j-th speaker was
    # present at frame i.
    onsets = [turn.onset for turn in turns]
    offsets = [turn.offset for turn in turns]
    speaker_ids = [turn.speaker_id for turn in turns]
    speaker_classes, speaker_class_inds = np.unique(
        speaker_ids, return_inverse=True)
    dur = max(score_offset for score_onset, score_offset in score_regions)
    n_frames = int(dur/step)
    X = np.zeros((n_frames, speaker_classes.size), dtype='int32')
    times = step*np.arange(n_frames)
    bis = np.searchsorted(times, onsets)
    eis = np.searchsorted(times, offsets)
    for bi, ei, speaker_class_ind in zip(bis, eis, speaker_class_inds):
        X[bi:ei, speaker_class_ind] = 1

    # Eliminate frames belonging to non-score regions.
    keep = np.zeros(len(X), dtype=bool)
    for score_onset, score_offset in score_regions:
        bi, ei = np.searchsorted(times, (score_onset, score_offset))
        keep[bi:ei] = True
    X = X[keep, ]

    return X


# TODO: Consider mapping all speech overlaps to a single class.
def flatten_labels(labels):
    """Helper function to convert output of ``turns_to_frames`` to 1-D array of
    unique values.

    Each row of ``labels`` is mapped to an integer representing an element of
    the powerset ``2**n_speakers``. The result is a 1-D array of integer labels
    in which each speaker, each possible overlap of speakers, and non-speech
    are differentiated. This is a necessary pre-processing step for the
    clustering metrics.

    Parameters
    ---------
    labels : ndarray, (n_frames, n_speakers)
        Frame-level speaker labels. The ``i,j``-th entry of this array is 1
        if speaker ``j``-th speaker was present at frame ``i`` and 0 otherwise.

    Returns
    -------
    flattened_labels : ndarray, (n_frames,)
        Flattened frame labels..
    """
    pows = 2**np.arange(labels.shape[1])
    flattened_labels = np.sum(pows*labels, axis=1)
    return flattened_labels


class Scores(namedtuple(
        'Scores',
        ['file_id', 'der', 'jer', 'bcubed_precision', 'bcubed_recall',
         'bcubed_f1', 'tau_ref_sys', 'tau_sys_ref', 'ce_ref_sys',
         'ce_sys_ref', 'mi', 'nmi'])):
    """Structure containing metrics.

    Parameters
    ----------
    file_id : str
        File id for file scored.

    der : float
        Diarization error rate in percent.

    jer : float
        Jaccard error rate in percent.

    bcubed_precision : float
        B-cubed precision.

    bcubed_recall : float
        B-cubed recall.

    bcubed_f1 : float
        B-cubed F1.

    tau_ref_sys : float
        Value between 0 and 1 that is high when the reference diarization is
        predictive of the system diarization and low when the reference
        diarization provides essentially no information about the system
        diarization.

    tau_sys_ref : float
        Value between 0 and 1 that is high when the system diarization is
        predictive of the reference diarization and low when the system
        diarization provides essentially no information about the reference
        diarization.

    ce_ref_sys : float
        Conditional entropy of the reference diarization given the system
        diarization.

    ce_sys_ref : float
        Conditional entropy of the system diarization given the reference
        diarization.

    mi : float
        Mutual information.

    nmi : float
        Normalized mutual information.
    """
    __slots__ = ()


def score(ref_turns, sys_turns, uem, step=0.010, nats=False, jer_min_ref_dur=0.0,
          **kwargs):
    """Score diarization.

    Parameters
    ----------
    ref_turns : list of Turn
        Reference speaker turns.

    sys_turns : list of Turn
        System speaker turns.

    uem : UEM
        Un-partitioned evaluation map.

    step : float, optional
        Frame step size  in seconds. Not relevant for computation of DER.
        (Default: 0.01)

    nats : bool, optional
        If True, use nats as unit for information theoretic metrics.
        Otherwise, use bits.
        (Default: False)

    jer_min_ref_dur : float, optional
        Minimum reference speaker duration in seconds for JER calculation.
        Reference speakers with durations less than ``min_ref_dur`` will be
        excluded for scoring purposes. Setting this to a small non-zero number
        may stabilize JER when the reference segmentation contains multiple
        extraneous speakers.
        (Default: 0.0)

    kwargs
        Keyword arguments to be passed to ``metrics.der``.

    Returns
    -------
    file_scores : list of Scores
        Scores for all files.

    global_scores : Scores
        Global scores.
    """
    if jer_min_ref_dur is not None:
        jer_min_ref_dur = int(jer_min_ref_dur/step)

    # Build contingency matrices.
    file_to_ref_turns = defaultdict(
        list,
        {fid : list(g) for fid, g in groupby(ref_turns, lambda x: x.file_id)})
    file_to_sys_turns =defaultdict(
        list,
        {fid : list(g) for fid, g in groupby(sys_turns, lambda x: x.file_id)})
    file_to_cm = {} # Map from files to contingency matrices used by
                    # clustering metrics.
    file_to_jer_cm = {} # Map from files to contingency matrices used by
                        # JER.
    file_to_ref_durs = {} # Map from files to speaker durations in reference
                          # segmentation.
    file_to_sys_durs = {} # Map from files to speaker durations in system
                          # segmentation.
    for file_id, score_regions in iteritems(uem):
        ref_labels = turns_to_frames(
            file_to_ref_turns[file_id], score_regions, step=step)
        sys_labels = turns_to_frames(
            file_to_sys_turns[file_id], score_regions, step=step)
        file_to_ref_durs[file_id] = ref_labels.sum(axis=0)
        file_to_sys_durs[file_id] = sys_labels.sum(axis=0)
        file_to_jer_cm[file_id] = metrics.contingency_matrix(
            ref_labels, sys_labels)
        file_to_cm[file_id] = metrics.contingency_matrix(
            flatten_labels(ref_labels), flatten_labels(sys_labels))
    global_cm = block_diag(*list(itervalues(file_to_cm)))
    # Above line has the undesirable property of claiming silence on
    # different files is a different category. However, leave it in for
    # consistency with how the clustering metrics were computed in DIHARD I.

    # Compute DER. This bit is slow as it relies on NIST's perl script.
    file_to_der, global_der = metrics.der(
        ref_turns, sys_turns, uem=uem, **kwargs)

    # Compute JER.
    file_to_jer, global_jer = metrics.jer(
        file_to_ref_durs, file_to_sys_durs, file_to_jer_cm, jer_min_ref_dur)

    # Compute clustering metrics.
    def compute_metrics(fid, cm, der, jer):
        bcubed_precision, bcubed_recall, bcubed_f1 = metrics.bcubed(
            None, None, cm)
        tau_ref_sys, tau_sys_ref = metrics.goodman_kruskal_tau(
            None, None, cm)
        ce_ref_sys = metrics.conditional_entropy(None, None, cm, nats)
        ce_sys_ref = metrics.conditional_entropy(None, None, cm.T, nats)
        mi, nmi = metrics.mutual_information(None, None, cm, nats)
        return Scores(
            fid, der, jer, bcubed_precision, bcubed_recall, bcubed_f1,
            tau_ref_sys, tau_sys_ref, ce_ref_sys, ce_sys_ref, mi, nmi)
    file_scores = []
    for file_id, cm in iteritems(file_to_cm):
        file_scores.append(compute_metrics(
            file_id, cm, file_to_der[file_id], jer=file_to_jer[file_id]))
    global_scores = compute_metrics(
        '*** OVERALL ***', global_cm, global_der, global_jer)

    return file_scores, global_scores



"""Tets for UEM utilities."""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import os
import shutil
import tempfile

from numpy.testing import (assert_equal, assert_raises, assert_raises_regex)
import pytest

from scorelib.turn import Turn
from scorelib.uem import gen_uem, load_uem, write_uem, UEM


TEST_DIR = os.path.dirname(os.path.abspath(__file__))
TMP_DIR = tempfile.mkdtemp(prefix="dscore_scorelib_test_uem__")


def test_UEM_input_validation():
    # Test type validation.
    invalid_type_msg = 'Expected sequence of pairs'
    with assert_raises_regex(TypeError, invalid_type_msg):
        UEM()['fid'] = 1
    with assert_raises_regex(TypeError, invalid_type_msg):
        UEM()['fid'] = (1, (2, 3))
    with assert_raises_regex(TypeError, invalid_type_msg):
        UEM()['fid'] = ('1', (2, 3))
    with assert_raises_regex(TypeError, invalid_type_msg):
        UEM()['fid'] = ((1,), (2, 3))
    with assert_raises_regex(TypeError, invalid_type_msg):
        UEM()['fid'] = ((3, 4, 5), (0, 1))

    # Test value validation.
    invalid_timestamp_msg = 'Could not convert interval'
    with assert_raises_regex(ValueError, invalid_timestamp_msg):
        UEM()['fid'] = (('a', 'b'), (2, 3))
    invalid_interval_msg = 'Invalid interval'
    with assert_raises_regex(ValueError, invalid_interval_msg):
        UEM()['fid'] = ((-1, 1), (2, 3))
    with assert_raises_regex(ValueError, invalid_interval_msg):
        UEM()['fid'] = ((1, 1), (2, 3))
    with assert_raises_regex(ValueError, invalid_interval_msg):
        UEM()['fid'] = ((-1, 1), (2, 3))


def test_UEM_merge_overlaps():
    overlaps_uem = UEM({'fid' : [(0, 5), (4, 10)]})
    no_overlaps_uem = UEM({'fid' : [(0, 10)]})
    assert overlaps_uem == no_overlaps_uem


def test_load_uem():
    expected_uem = UEM({
        'FILE1' : [(0, 15), (25, 30.4)],
        'FILE2' : [(4.5, 13.24)]})
    loaded_uem = load_uem(os.path.join(TEST_DIR, 'test_load.uem'))
    assert expected_uem == loaded_uem


def test_write_uem():
    uem = UEM({
        'FILE1' : [(0, 5), (4, 10.6)],
        'FILE2' : [(4.5, 13.24)]})
    tmp_uemf = os.path.join(TMP_DIR, 'test_write.uem')
    write_uem(tmp_uemf, uem)
    expected_uemf = os.path.join(TEST_DIR, 'test_write.uem')
    with open(tmp_uemf, 'rb') as f, open(expected_uemf, 'rb') as g:
        assert f.read() == g.read()


def test_gen_uem():
    ref_turns = [
        Turn(0, 14, speaker_id='SPK1', file_id='FILE1'),
        Turn(20, 25, speaker_id='SPK2', file_id='FILE1'),
        Turn(11, 45, speaker_id='SPK3', file_id='FILE2'),
    ]
    sys_turns = [
        Turn(35, 46, speaker_id='SPK1', file_id='FILE1'),
        Turn(2.5, 14, speaker_id='SPK4', file_id='FILE2'),
    ]
    expected_uem = UEM({
        'FILE1' : [(0, 46)],
        'FILE2' : [(2.5, 45)]})
    assert expected_uem == gen_uem(ref_turns, sys_turns)


def teardown_module():
    if os.path.isdir(TMP_DIR):
        shutil.rmtree(TMP_DIR)



"""Tests for scoring module."""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import os

import numpy as np
from numpy.testing import (assert_almost_equal, assert_equal, assert_raises_regex)

from scorelib.rttm import load_rttm
from scorelib.score import flatten_labels, score, turns_to_frames, Scores
from scorelib.turn import Turn
from scorelib.uem import UEM


TEST_DIR = os.path.dirname(os.path.abspath(__file__))


def test_turns_to_frames():
    # Check validation.
    turns = [
        Turn(0, 11, speaker_id='S1', file_id='FILE1'),
        Turn(5, 11, speaker_id='S2', file_id='FILE2'),
        ]
    with assert_raises_regex(ValueError, 'Turns should be from'):
        labels = turns_to_frames(turns, [(0, 1)], step=0.1)

    # Check that file containing no speech returns an (n_frames, 0) array.
    labels = turns_to_frames([], [(0, 1)], step=0.1)
    assert_equal(labels, np.zeros((10, 0), dtype='int64'))

    # Check on toy input.
    expected_labels = np.zeros((120, 2), dtype='int64')
    expected_labels[:110, 0] = 1
    expected_labels[50:110, 1] = 1
    turns = [
        Turn(0, 11, speaker_id='S1', file_id='FILE1'),
        Turn(5, 11, speaker_id='S2', file_id='FILE1'),
        ]
    labels = turns_to_frames(turns, [(0, 12)], step=0.1)
    assert_equal(labels, expected_labels)


def test_flatten_labels():
    # No speech.
    assert_equal(flatten_labels(np.zeros((5, 0), dtype='int64')),
                 np.zeros(5, dtype='int64'))
    assert_equal(flatten_labels(np.zeros((5, 1), dtype='int64')),
                 np.zeros(5, dtype='int64'))
    assert_equal(flatten_labels(np.zeros((5, 2), dtype='int64')),
                 np.zeros(5, dtype='int64'))

    # Toy input with 2 speakers.
    labels = np.array(
        [[0, 0],
         [1, 0],
         [0, 1],
         [1, 1]],
        dtype='int64')
    assert_equal(flatten_labels(labels),
                 np.arange(4, dtype='int64'))


def test_score():
    # Some real data.
    expected_scores = Scores(
        'FILE1', 26.39309, 33.24631, 0.71880, 0.72958, 0.72415, 0.60075, 0.58534,
        0.80471, 0.72543, 0.96810, 0.55872)
    ref_turns, _, _ = load_rttm(os.path.join(TEST_DIR, 'ref.rttm'))
    sys_turns, _, _ = load_rttm(os.path.join(TEST_DIR, 'sys.rttm'))
    uem = UEM({'FILE1' : [(0, 43)]})
    file_scores, global_scores = score(ref_turns, sys_turns, uem)
    assert len(file_scores) == 1
    assert file_scores[-1].file_id == expected_scores.file_id
    assert_almost_equal(file_scores[-1][1:], expected_scores[1:], 3)
    expected_scores = expected_scores._replace(file_id='*** OVERALL ***')
    assert global_scores.file_id == expected_scores.file_id
    assert_almost_equal(global_scores[1:], expected_scores[1:], 3)



from typing import List

import numpy as np
from .asteroid_metric import get_metrics

SAMPLE_RATE = 16000


def pesq(preds: List[np.array], refs: List[np.array]):
    """
    Args:
        preds (list): a list of waveform for reconstructed prediction
        refs (list): a list of waveform for clean reference

    Return: float
        Each waveform pair get a pesq value.
        The final score is the average over all pesq values.
    """
    pesqs = []
    for pred, ref in zip(preds, refs):
        metrics = get_metrics(
            pred, ref, pred, sample_rate=SAMPLE_RATE, metrics_list=["pesq"]
        )
        pesqs.append(metrics["pesq"])
    return sum(pesqs) / len(pesqs)


def stoi(preds: List[np.array], refs: List[np.array]):
    """
    Args:
        preds (list): a list of waveform for reconstructed prediction
        refs (list): a list of waveform for clean reference

    Return: float
        Each waveform pair get a stoi value.
        The final score is the average over all stoi values.
    """
    stois = []
    for pred, ref in zip(preds, refs):
        metrics = get_metrics(
            pred, ref, pred, sample_rate=SAMPLE_RATE, metrics_list=["stoi"]
        )
        stois.append(metrics["stoi"])
    return sum(stois) / len(stois)


def sisdri(mixtures: List[np.array], preds: List[np.array], refs: List[np.array]):
    """
    SISDRi for source separation. It measure the SISDR improvement from 'input noisy mixure v.s. clean reference'
    to 'output prediction v.s. clean reference'. The postfix 'i' stands for 'improvement'

    Args:
        mixtures (list): Each element is a np.array in the shape of (n_samples, )
        preds (list): Each pred is a np.array in the shape of (n_source, n_samples)
        refs (list): Each ref is a np.array in the shape of (n_source, n_samples)

    Return: float
        The corpus-level SISDRi. Each pred and ref pair get a SISDRi.
        The final score is the corpus-level average.
    """
    sisdris = []
    for mixture, pred, ref in zip(mixtures, preds, refs):
        metrics = get_metrics(
            mixture[None, :],
            ref,
            pred,
            sample_rate=SAMPLE_RATE,
            metrics_list=["si_sdr"],
            compute_permutation=True,
        )
        sisdri = metrics["si_sdr"] - metrics["input_si_sdr"]
        sisdris.append(sisdri)
    return sum(sisdris) / len(sisdris)



"""Tets for UEM utilities."""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from intervaltree import Interval, IntervalTree

import pytest

from scorelib.turn import chop_tree, merge_turns, trim_turns, Turn
from scorelib.uem import UEM


def test_merge_turns():
    expected_turns = [
        Turn(0, 11, speaker_id='S1', file_id='FILE1'),
        Turn(0, 11, speaker_id='S2', file_id='FILE1'),
        Turn(0, 11, speaker_id='S1', file_id='FILE2'),
        ]
    turns = [
        Turn(0, 10, speaker_id='S1', file_id='FILE1'),
        Turn(9, 11, speaker_id='S1', file_id='FILE1'),
        Turn(0, 11, speaker_id='S2', file_id='FILE1'),
        Turn(0, 11, speaker_id='S1', file_id='FILE2'),
        ]
    assert set(expected_turns) == set(merge_turns(turns))


def test_trim_turns():
    turns = [
        Turn(1, 5, speaker_id='S1', file_id='FILE1'),
        Turn(6, 10, speaker_id='S1', file_id='FILE1'),
        Turn(0, 10, speaker_id='S1', file_id='FILE2'),
        ]

    # Trim with a UEM.
    uem = UEM({'FILE1' : [(2, 6), (5.8, 7)],
               'FILE2' : [(2, 3), (4, 5)]})
    expected_turns = [
        Turn(2, 5, speaker_id='S1', file_id='FILE1'),
        Turn(6, 7, speaker_id='S1', file_id='FILE1'),
        Turn(2, 3, speaker_id='S1', file_id='FILE2'),
        Turn(4, 5, speaker_id='S1', file_id='FILE2'),
        ]
    assert set(expected_turns) == set(trim_turns(turns, uem))

    # Trim without UEM.
    expected_turns = [
        Turn(2, 5, speaker_id='S1', file_id='FILE1'),
        Turn(6, 7, speaker_id='S1', file_id='FILE1'),
        Turn(2, 7, speaker_id='S1', file_id='FILE2'),
        ]
    assert set(expected_turns) == set(trim_turns(turns, None, 2, 7))


def test_chop_tree():
    def _get_tree():
        return IntervalTree.from_tuples(
            [(1, 5, 'i1'),
             (7, 10, 'i2'),
            ])

    # Interval contained within chop region.
    expected_tree = IntervalTree.from_tuples([(7, 10, 'i2')])
    tree = _get_tree()
    overlap_intervals = chop_tree(tree, 0, 6)
    assert overlap_intervals == set([Interval(1, 5, 'i1')])
    assert tree == expected_tree

    # Left overlap.
    expected_tree = IntervalTree.from_tuples([(2, 5, 'i1'), (7, 10, 'i2')])
    tree = _get_tree()
    overlap_intervals = chop_tree(tree, 0, 2)
    assert overlap_intervals == set([Interval(1, 5, 'i1')])
    assert tree == expected_tree

    # Right overlap.
    expected_tree = IntervalTree.from_tuples([(1, 5, 'i1'), (7, 9, 'i2')])
    tree = _get_tree()
    overlap_intervals = chop_tree(tree, 9, 11)
    assert overlap_intervals == set([Interval(7, 10, 'i2')])
    assert tree == expected_tree

    # Chop region contained within interval.
    expected_tree = IntervalTree.from_tuples([(1, 2, 'i1'),
                                              (3, 5, 'i1'),
                                              (7, 10, 'i2')])
    tree = _get_tree()
    overlap_intervals = chop_tree(tree, 2, 3)
    assert overlap_intervals == set([Interval(1, 5, 'i1')])
    assert tree == expected_tree

    # Overlaps two intervals.
    expected_tree = IntervalTree.from_tuples([(1, 4, 'i1'),
                                              (8, 10, 'i2')])
    tree = _get_tree()
    overlap_intervals = chop_tree(tree, 4, 8)
    assert overlap_intervals == set([Interval(1, 5, 'i1'),
                                     Interval(7, 10, 'i2')])
    assert tree == expected_tree

    # No overlap.
    expected_tree = _get_tree()
    tree = _get_tree()
    overlap_intervals = chop_tree(tree, 6, 6.5)
    assert overlap_intervals == set()
    assert tree == expected_tree

    # No trivial overlaps.
    expected_tree = _get_tree()
    tree = _get_tree()
    overlap_intervals = chop_tree(tree, 0, 1)
    assert overlap_intervals == set()
    assert tree == expected_tree
    overlap_intervals = chop_tree(tree, 10, 11)
    assert overlap_intervals == set()
    assert tree == expected_tree



"""Functions for scoring frame-level diarization output."""
# TODO: Module is too long. Refactor.
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import os
import re
import shutil
import subprocess
import tempfile

import numpy as np
from scipy.optimize import linear_sum_assignment
from scipy.sparse import coo_matrix, issparse

from .rttm import write_rttm
from .uem import gen_uem, write_uem
from .utils import clip, xor

__all__ = ['bcubed', 'conditional_entropy', 'contingency_matrix', 'der',
           'goodman_kruskal_tau', 'jer', 'mutual_information']


EPS = np.finfo(float).eps


def contingency_matrix(ref_labels, sys_labels):
    """Return contingency matrix between ``ref_labels`` and ``sys_labels``.

    Parameters
    ----------
    ref_labels : ndarray, (n_samples,) or (n_samples, n_ref_classes)
        Reference labels encoded using one-hot scheme.

    sys_labels : ndarray, (n_samples,) or ((n_samples, n_sys_classes)
        System labels encoded using one-hot scheme.

    Returns
    -------
    cm : ndarray, (n_ref_classes, n_sys_classes)
        Contigency matrix whose ``i, j``-th entry is the number of times the
        ``i``-th reference label and ``j``-th system label co-occur.
    """
    if ref_labels.ndim != sys_labels.ndim:
        raise ValueError(
            'ref_labels and sys_labels should either both be 1D arrays of '
            'labels or both be 2D arrays of one-hot encoded labels: shapes '
            'are %r, %r' % (ref_labels.shape, sys_labels.shape))
    if ref_labels.shape[0] != sys_labels.shape[0]:
        raise ValueError(
            'ref_labels and sys_labels must have same size: received %d '
            'and %d' % (ref_labels.shape[0], sys_labels.shape[0]))
    if ref_labels.ndim == 1:
        ref_classes, ref_class_inds = np.unique(
            ref_labels, return_inverse=True)
        sys_classes, sys_class_inds = np.unique(
            sys_labels, return_inverse=True)
        n_frames = ref_labels.size
        cm = coo_matrix(
            (np.ones(n_frames), (ref_class_inds, sys_class_inds)),
            shape=(ref_classes.size, sys_classes.size),
            dtype=np.int64)
        cm = cm.toarray()
    else:
        ref_labels = ref_labels.astype('int64', copy=False)
        sys_labels = sys_labels.astype('int64', copy=False)
        cm = ref_labels.T.dot(sys_labels)
        if issparse(cm):
            cm = cm.toarray()
    return cm


def bcubed(ref_labels, sys_labels, cm=None):
    """Return B-cubed precision, recall, and F1.

    The B-cubed precision of an item is the proportion of items with its
    system label that share its reference label (Bagga and Baldwin, 1998).
    Similarly, the B-cubed recall of an item is the proportion of items
    with its reference label that share its system label. The overall B-cubed
    precision and recall, then, are the means of the precision and recall for
    each item.

    Parameters
    ----------
    ref_labels : ndarray, (n_frames,)
        Reference labels.

    sys_labels : ndarray, (n_frames,)
        System labels.

    cm : ndarray, (n_ref_classes, n_sys_classes)
        Contingency matrix between reference and system labelings. If None,
        will be computed automatically from ``ref_labels`` and ``sys_labels``.
        Otherwise, the given value will be used and ``ref_labels`` and
        ``sys_labels`` ignored.
        (Default: None)

    Returns
    -------
    precision : float
        B-cubed precision.

    recall : float
        B-cubed recall.

    f1 : float
        B-cubed F1.

    References
    ----------
    Bagga, A. and Baldwin, B. (1998). "Algorithms for scoring coreference
    chains." Proceedings of LREC 1998.
    """
    if cm is None:
        cm = contingency_matrix(ref_labels, sys_labels)
    cm = cm.astype('float64')
    cm_norm = cm / cm.sum()
    precision = np.sum(cm_norm * (cm / cm.sum(axis=0)))
    recall = np.sum(cm_norm * (cm / np.expand_dims(cm.sum(axis=1), 1)))
    f1 = 2*(precision*recall)/(precision + recall)
    return precision, recall, f1


def goodman_kruskal_tau(ref_labels, sys_labels, cm=None):
    """Return Goodman-Kruskal tau between ``ref_labels`` and ``sys_labels``.

    Parameters
    ----------
    ref_labels : ndarray, (n_frames,)
        Reference labels.

    sys_labels : ndarray, (n_frames,)
        System labels.

    cm : ndarray, (n_ref_classes, n_sys_classes)
        Contingency matrix between reference and system labelings. If None,
        will be computed automatically from ``ref_labels`` and ``sys_labels``.
        Otherwise, the given value will be used and ``ref_labels`` and
        ``sys_labels`` ignored.
        (Default: None)

    Returns
    -------
    tau_ref_sys : float
        Value between 0 and 1 that is high when ``ref_labels`` is predictive
        of ``sys_labels`` and low when ``ref_labels`` provides essentially no
        information about ``sys_labels``.

    tau_sys_ref : float
        Value between 0 and 1 that is high when ``sys_labels`` is predictive
        of ``ref_labels`` and low when ``sys_labels`` provides essentially no
        information about ``ref_labels``.

    References
    ----------
    - Goodman, L.A. and Kruskal, W.H. (1954). "Measures of association for
      cross classifications." Journal of the American Statistical Association.
    - Pearson, R. (2016). GoodmanKruskal: Association Analysis for Categorical
      Variables. https://CRAN.R-project.org/package=GoodmanKruskal.
    """
    if cm is None:
        cm = contingency_matrix(ref_labels, sys_labels)
    cm = cm.astype('float64')
    cm = cm / cm.sum()
    ref_marginals = cm.sum(axis=1)
    sys_marginals = cm.sum(axis=0)
    n_ref_classes, n_sys_classes = cm.shape

    # Tau(ref, sys).
    if n_sys_classes == 1:
        # Special case: only single class in system labeling, so any
        #               reference labeling is perfectly predictive.
        tau_ref_sys = 1.
    else:
        vy = 1 - np.sum(sys_marginals**2)
        xy_term = np.sum(cm**2, axis=1)
        vy_bar_x = 1 - np.sum(xy_term / ref_marginals)
        tau_ref_sys = (vy - vy_bar_x) / vy

    # Tau(sys, ref).
    if n_ref_classes == 1:
        # Special case: only single class in reference labeling, so any
        #               system labeling is perfectly predictive.
        tau_sys_ref = 1.
    else:
        vx = 1 - np.sum(ref_marginals**2)
        yx_term = np.sum(cm**2, axis=0)
        vx_bar_y = 1 - np.sum(yx_term / sys_marginals)
        tau_sys_ref = (vx - vx_bar_y) / vx

    return tau_ref_sys, tau_sys_ref


def conditional_entropy(ref_labels, sys_labels, cm=None, nats=False):
    """Return conditional entropy of ``ref_labels`` given ``sys_labels``.

    The conditional entropy ``H(ref | sys)`` quantifies how much information
    is needed to describe the reference labeling given that the system labeling
    is known. It is 0 when the labelings are identical and increases as the
    system labeling becomes less descriptive of the reference labeling.

    Parameters
    ----------
    ref_labels : ndarray, (n_frames,)
        Reference labels.

    sys_labels : ndarray, (n_frames,)
        System labels.

    cm : ndarray, (n_ref_classes, n_sys_classes)
        Contingency matrix between reference and system labelings. If None,
        will be computed automatically from ``ref_labels`` and ``sys_labels``.
        Otherwise, the given value will be used and ``ref_labels`` and
        ``sys_labels`` ignored.
        (Default: None)

    nats : bool, optional
        If True, return conditional entropy in nats. Otherwise, return in bits.
        (Default: False)

    References
    ----------
    - https://en.wikipedia.org/wiki/Conditional_entropy
    - Cover, T.M. and Thomas, J.A. (1991). Elements of Information Theory.
    - Rosenberg, A. and Hirschberg, J. (2007). "V-Measure: A conditional
      entropy-based external cluster evaluation measure." Proceedings of EMNLP
      2007.
    """
    log = np.log if nats else np.log2
    if cm is None:
        cm = contingency_matrix(ref_labels, sys_labels)
    sys_marginals = cm.sum(axis=0)
    N = cm.sum()
    ref_inds, sys_inds = np.nonzero(cm)
    vals = cm[ref_inds, sys_inds] # Non-zero values of contingency matrix.
    sys_marginals = sys_marginals[sys_inds] # Corresponding marginals.
    sigma = vals/N * (log(sys_marginals) - log(vals))
    return sigma.sum()


VALID_NORM_METHODS = set(['min', 'sum', 'sqrt', 'max'])

def mutual_information(ref_labels, sys_labels, cm=None, nats=False,
                       norm_method='sqrt'):
    """Return mutual information between ``ref_labels`` and ``sys_labels``.

    The mutual information ``I(ref, sys)`` quantifies how much information is
    shared by the reference and system labelings; that is, how much knowing
    one labeling reduces uncertainty about the other. It is 0 in the case that
    the labelings are independent and increases as they become more predictive
    of each other with a least upper bound of ``min(H(ref), H(sys))``.

    Normalized mutual information converts mutual information into a similarity
    metric ranging on [0, 1]. Multiple normalization schemes are available,
    set by the ``norm_method`` argument, which takes the following values:

    - ``min``  --  normalize by ``min(H(ref), H(sys))``
    - ``sum``  --  normalize by ``0.5*(H(ref) + H(sys))``
    - ``sqrt``  --  normalize by ``sqrt(H(ref)*H(sys))``
    - ``max``  --  normalize by ``max(H(ref), H(sys))``

    Parameters
    ----------
    ref_labels : ndarray, (n_frames,)
        Reference labels.

    sys_labels : ndarray, (n_frames,)
        System labels.

    cm : ndarray, (n_ref_classes, n_sys_classes)
        Contingency matrix between reference and system labelings. If None,
        will be computed automatically from ``ref_labels`` and ``sys_labels``.
        Otherwise, the given value will be used and ``ref_labels`` and
        ``sys_labels`` ignored.
        (Default: None)

    nats : bool, optional
        If True, return nats. Otherwise, return bits.
        (Default: False)

    norm_method : str, optional
        Normalization method for NMI computation.
        (Default: 'sqrt')

    Returns
    -------
    mi : float
        Mutual information.

    nmi : float
        Normalized mutual information.

    References
    ----------
    - https://en.wikipedia.org/wiki/Mutual_information
    - Cover, T.M. and Thomas, J.A. (1991). Elements of Information Theory.
    - Strehl, A. and Ghosh, J. (2002). "Cluster ensembles  -- A knowledge
      reuse framework for combining multiple partitions." Journal of Machine
      Learning Research.
    - Nguyen, X.V., Epps, J., and Bailey, J. (2010). "Information theoretic
      measures for clustering comparison: Variants, properties, normalization
      and correction for chance." Journal of Machine Learning Research.
    """
    if norm_method not in VALID_NORM_METHODS:
        raise ValueError('"%s" is not a valid NMI normalization method.')
    log = np.log if nats else np.log2
    if cm is None:
        cm = contingency_matrix(ref_labels, sys_labels)

    # Special cases in which one or more of H(ref) and H(sys) is
    # 0.
    n_ref_classes, n_sys_classes = cm.shape
    if xor(n_ref_classes == 1, n_sys_classes == 1):
        # Case 1: MI is by definition 0 as should be NMI, regardless of
        #         normalization.
        return 0.0, 0.0
    if n_ref_classes == n_sys_classes == 1:
        # Case 2: MI is 0, but as the data is not split, each clustering
        #         is perfectly predictive of the other, so set NMI to 1.
        return 0.0, 1.0

    # Mutual information.
    N = cm.sum()
    ref_marginals = cm.sum(axis=1)
    sys_marginals = cm.sum(axis=0)
    ref_inds, sys_inds = np.nonzero(cm)
    vals = cm[ref_inds, sys_inds] # Non-zero values of contingency matrix.
    outer = ref_marginals[ref_inds]*sys_marginals[sys_inds]
    sigma = (vals/N) * (
        log(vals) - log(outer) + log(N))
    mi = sigma.sum()
    mi = max(mi, 0.)

    # Normalized mutual information.
    def h(p):
        p = p[p > 0]
        return max(-np.sum(p*log(p)), 0)
    h_ref = h(ref_marginals / N)
    h_sys = h(sys_marginals / N)
    if norm_method == 'max':
        denom = max(h_ref, h_sys)
    elif norm_method == 'sum':
        denom = 0.5*(h_ref + h_sys)
    elif norm_method == 'sqrt':
        denom = np.sqrt(h_ref*h_sys)
    elif norm_method == 'min':
        denom = min(h_ref, h_sys)
    nmi = mi / denom
    nmi = clip(nmi, 0., 1.)

    return mi, nmi


SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))
MDEVAL_BIN = os.path.join(SCRIPT_DIR, 'md-eval-22.pl')
FILE_REO = re.compile(r'(?<=Speaker Diarization for).+(?=\*\*\*)')
SCORED_SPEAKER_REO = re.compile(r'(?<=SCORED SPEAKER TIME =)[\d.]+')
MISS_SPEAKER_REO = re.compile(r'(?<=MISSED SPEAKER TIME =)[\d.]+')
FA_SPEAKER_REO = re.compile(r'(?<=FALARM SPEAKER TIME =)[\d.]+')
ERROR_SPEAKER_REO = re.compile(r'(?<=SPEAKER ERROR TIME =)[\d.]+')

# TODO: Working with md-eval is a PITA, even with modifications to the
#       reporting. Suggest looking into moving over to pyannote's
#       implementation.
def der(ref_turns, sys_turns, collar=0.0, ignore_overlaps=False, uem=None):
    """Return overall diarization error rate.

    Diarization error rate (DER), introduced for the NIST Rich Transcription
    evaluations, is computed as the sum of the following:

    - speaker error  --  percentage of scored time for which the wrong speaker
      id is assigned within a speech region
    - false alarm speech  --   percentage of scored time for which a nonspeech
      region is incorrectly marked as containing speech
    - missed speech  --  percentage of scored time for which a speech region is
      incorrectly marked as not containing speech

    As with word error rate, a score of zero indicates perfect performance and
    higher scores (which may exceed 100) indicate poorer performance.

    DER is computed as defined in the NIST RT-09 evaluation plan using version
    22 of the ``md-eval.pl`` scoring script. When ``ignore_overlaps=False``,
    this is equivalent to running the following command:

        md-eval.pl -r ref.rttm -s sys.rttm -c collar -u uemf

    where ``ref.rttm`` and ``sys.rttm`` are RTTM files produced from
    ``ref_turns`` and ``sys_turns`` respectively and ``uemf`` is an
    Un-partitioned Evaluation Map (UEM) file delimiting the scoring regions.
    If a ``UEM`` instance is supplied via the``uem`` argument, this file will
    be created from the supplied UEM. Otherwise, it will be generated
    automatically from ``ref_turns`` and ``sys_turns`` using the
    ``uem.gen_uem`` function. Similarly, when ``ignore_overlaps=True``:

        md-eval.pl -r ref.rttm -s sys.rttm -c collar -u uemf -1

    Parameters
    ----------
    ref_turns : list of Turn
        Reference speaker turns.

    sys_turns : list of Turn
        System speaker turns.

    collar : float, optional
        Size of forgiveness collar in seconds. Diarization output will not be
        evaluated within +/- ``collar`` seconds of reference speaker
        boundaries.
        (Default: 0.0)

    ignore_overlaps : bool, optional
        If True, ignore regions in the reference diarization in which more
        than one speaker is speaking.
        (Default: False)

    uem : UEM, optional
        Evaluation map. If not supplied, will be generated automatically from
        ``ref_turns`` and ``sys_turns``.
        (Default: None)

    Returns
    -------
    file_to_der : dict
        Mapping from files to diarization error rates (in percent) for those
        files.

    global_der : float
        Overall diarization error rate (in percent).

    References
    ----------
    NIST. (2009). The 2009 (RT-09) Rich Transcription Meeting Recognition
    Evaluation Plan. https://web.archive.org/web/20100606041157if_/http://www.itl.nist.gov/iad/mig/tests/rt/2009/docs/rt09-meeting-eval-plan-v2.pdf
    """
    tmp_dir = tempfile.mkdtemp()

    # Write RTTMs.
    ref_rttm_fn = os.path.join(tmp_dir, 'ref.rttm')
    write_rttm(ref_rttm_fn, ref_turns)
    sys_rttm_fn = os.path.join(tmp_dir, 'sys.rttm')
    write_rttm(sys_rttm_fn, sys_turns)

    # Write UEM.
    if uem is None:
        uem = gen_uem(ref_turns, sys_turns)
    uemf = os.path.join(tmp_dir, 'all.uem')
    write_uem(uemf, uem)

    # Actually score.
    try:
        cmd = [MDEVAL_BIN,
               '-af',
               '-r', ref_rttm_fn,
               '-s', sys_rttm_fn,
               '-c', str(collar),
               '-u', uemf,
              ]
        if ignore_overlaps:
            cmd.append('-1')
        stdout = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        stdout = e.output
    finally:
        shutil.rmtree(tmp_dir)

    # Parse md-eval output to extract by-file and total scores.
    stdout = stdout.decode('utf-8')
    file_ids = [m.strip() for m in FILE_REO.findall(stdout)]
    file_ids = [file_id[2:] if file_id.startswith('f=') else file_id
                for file_id in file_ids]
    scored_speaker_times = np.array(
        [float(m) for m in SCORED_SPEAKER_REO.findall(stdout)])
    miss_speaker_times = np.array(
        [float(m) for m in MISS_SPEAKER_REO.findall(stdout)])
    fa_speaker_times = np.array(
        [float(m) for m in FA_SPEAKER_REO.findall(stdout)])
    error_speaker_times = np.array(
        [float(m) for m in ERROR_SPEAKER_REO.findall(stdout)])
    with np.errstate(invalid='ignore', divide='ignore'):
        error_times = miss_speaker_times + fa_speaker_times + error_speaker_times
        ders = error_times / scored_speaker_times
    ders[np.isnan(ders)] = 0 # Numerator and denominator both 0.
    ders[np.isinf(ders)] = 1 # Numerator > 0, but denominator = 0.
    ders *= 100. # Convert to percent.

    # Reconcile with UEM, keeping in mind that in the edge case where no
    # reference turns are observed for a file, md-eval doesn't report results
    # for said file.
    file_to_der_base = dict(zip(file_ids, ders))
    file_to_der = {}
    for file_id in uem:
        try:
            der = file_to_der_base[file_id]
        except KeyError:
            # Check for any system turns for that file, which should be FAs,
            # assuming that the turns have been cropped to the UEM scoring
            # regions.
            n_sys_turns = len(
                [turn for turn in sys_turns if turn.file_id == file_id])
            der = 100. if n_sys_turns else 0.0
        file_to_der[file_id] = der
    global_der = file_to_der_base['ALL']

    return file_to_der, global_der


def jer(file_to_ref_durs, file_to_sys_durs, file_to_cm, min_ref_dur=0):
    """Return Jacard error rate.

    Jaccard error rate (JER) rate is based on the Jaccard index, a similarity
    measure used to evaluate the output of image segmentation systems. An
    optimal mapping between reference and system speakers is determined and
    for each pair the Jaccard index is computed. The Jaccard error rate is then
    defined as 1 minus the average of these scores.

    More concretely, assume we have ``N`` reference speakers and ``M`` system
    speakers. An optimal mapping between speakers is determined using the
    Hungarian algorithm so that each reference speaker is paired with at most
    one system speaker and each system speaker with at most one reference
    speaker. Then, for each reference speaker ``ref`` the speaker-specific
    Jaccard error rate is ``(FA + MISS)/TOTAL``, where:
    - ``TOTAL`` is the duration of the union of reference and system speaker
      segments; if the reference speaker was not paired with a system speaker,
      it is the duration of all reference speaker segments
    - ``FA`` is the total system speaker time not attributed to the reference
      speaker; if the reference speaker was not paired with a system speaker,
      it is 0
    - ``MISS`` is the total reference speaker time not attributed to the
      system speaker; if the reference speaker was not paired with a system
      speaker, it is equal to ``TOTAL``
    The Jaccard error rate then is the average of the speaker specific Jaccard
    error rates.

    JER and DER are highly correlated with JER typically being higher, especially
    in recordings where one or more speakers is particularly dominant. Where it
    tends to track DER is in outliers where the diarization is especially bad,
    resulting on one or more unmapped system speakers whose speech is not then
    penalized. In these cases, where DER can easily exceed 500%, JER will never
    exceed 100% and may be far lower if the reference speakers are handled
    correctly. For this reason, it may be useful to pair JER with another metric
    evaluating speech detection and/or speaker overlap detection.

    Parameters
    ----------
    file_to_ref_durs : dict
        Mapping from files to durations of reference speakers in those files.

    file_to_sys_durs : dict
        Mapping from files to durations of system speakers in those files.

    file_to_cm : dict
        Mapping from files to contingency matrices for speakers in those files.

    min_ref_dur : float, optional
        Minimum reference speaker duration. Reference speakers with durations
        less than ``min_ref_dur`` will be excluded for scoring purposes. Setting
        this to a small non-zero number may stabilize JER when the reference
        segmentation contains multiple extraneous speakers.
        (Default: 0.0)

    Returns
    -------
    file_to_jer : dict
        Mapping from files to Jaccard error rates (in percent) for those files.

    global_jer : float
        Overall Jaccard error rate (in percent).

    References
    ----------
    https://en.wikipedia.org/wiki/Jaccard_index
    """
    # TODO: Explore treating non-speech as additional speaker for computation to
    #       more gracefully deal with exceptionally poor system performance.
    ref_dur_fids = set(file_to_ref_durs.keys())
    sys_dur_fids = set(file_to_sys_durs.keys())
    cm_fids = set(file_to_cm.keys())
    if not ref_dur_fids == sys_dur_fids == cm_fids:
        raise ValueError(
            'All passed dicts must have same keys.')
    file_ids = ref_dur_fids
    file_to_jer = {}
    all_speaker_jers = []
    n_ref_speakers_global = 0
    n_sys_speakers_global = 0
    for file_id in file_ids:
        # Filter.
        ref_durs = file_to_ref_durs[file_id]
        sys_durs = file_to_sys_durs[file_id]
        cm = file_to_cm[file_id]
        ref_keep = ref_durs >= min_ref_dur
        ref_durs = ref_durs[ref_keep]
        cm = cm[ref_keep, ]
        n_ref_speakers = ref_durs.size
        n_sys_speakers = sys_durs.size
        n_ref_speakers_global += n_ref_speakers
        n_sys_speakers_global += n_sys_speakers

        # Handle edge cases where either reference or system segmentation
        # posited no speech.
        if n_ref_speakers == 0 and n_sys_speakers > 0:
            # Case 1: no reference speech.
            file_to_jer[file_id] = 100.0
            continue
        elif n_ref_speakers > 0 and n_sys_speakers == 0:
            # Case 2: no system speech.
            file_to_jer[file_id] = 100.0
            all_speaker_jers.extend([100.]*n_ref_speakers)
            continue
        elif n_ref_speakers == 0 and n_sys_speakers == 0:
            # Case 3: no reference or system speech
            file_to_jer[file_id] = 0.0
            continue

        # Determine all speaker-level JER.
        ref_durs = np.tile(ref_durs, [n_sys_speakers, 1]).T
        sys_durs = np.tile(sys_durs, [n_ref_speakers, 1])
        intersect = cm
        union = ref_durs + sys_durs - intersect
        jer_speaker = 1 - intersect / union

        # Find dominant mapping by Hungarian algorithm (scipy >= 0.17) and compute
        # JER.
        ref_speaker_inds, sys_speaker_inds = linear_sum_assignment(jer_speaker)
        jers = np.ones(n_ref_speakers, dtype='float64')
        for ref_speaker_ind, sys_speaker_ind in zip(
                ref_speaker_inds, sys_speaker_inds):
            jers[ref_speaker_ind] = jer_speaker[ref_speaker_ind,
                                                sys_speaker_ind]
        jers *= 100.
        file_to_jer[file_id] = jers.mean()
        all_speaker_jers.extend(jers)

    # Determine global JER.
    if n_ref_speakers_global == 0 and n_sys_speakers_global > 0:
        # Case 1: no reference speech on ANY file.
        global_jer = 100.
    elif n_ref_speakers_global > 0 and n_sys_speakers_global == 0:
        # Case 2: no system speech on ANY file.
        global_jer = 100.
    elif n_ref_speakers_global == n_sys_speakers_global == 0:
        # Case 3: no reference OR system speech on ANY file.
        global_jer = 0.0
    else:
        # General case: at least 1 reference and 1 system speaker present.
        global_jer = np.mean(all_speaker_jers)

    return file_to_jer, global_jer



"""
Modified from https://github.com/nryant/dscore/blob/master/score.py
Original author: Neville Ryant
Modified author: Shu-wen Yang
"""

import re
import sys
from typing import List, Tuple

from .dscorelib.turn import Turn, merge_turns, trim_turns
from .dscorelib.score import score
from .dscorelib.six import iterkeys
from .dscorelib.uem import gen_uem
from .dscorelib.utils import info, warn


def check_for_empty_files(ref_turns, sys_turns, uem):
    """Warn on files in UEM without reference or speaker turns."""
    ref_file_ids = {turn.file_id for turn in ref_turns}
    sys_file_ids = {turn.file_id for turn in sys_turns}
    for file_id in sorted(iterkeys(uem)):
        if file_id not in ref_file_ids:
            warn('File "%s" missing in reference RTTMs.' % file_id)
        if file_id not in sys_file_ids:
            warn('File "%s" missing in system RTTMs.' % file_id)
    # TODO: Clarify below warnings; this indicates that there are no
    #       ELIGIBLE reference/system turns.
    if not ref_turns:
        warn("No reference speaker turns found within UEM scoring regions.")
    if not sys_turns:
        warn("No system speaker turns found within UEM scoring regions.")


def der(
    hyp_turns: List[Tuple[float, float, str, str]],
    ref_turns: List[Tuple[float, float, str, str]],
    step: float = 0.010,
    jer_min_ref_dur: float = 0.0,
    collar: float = 0.0,
    ignore_overlaps: bool = False,
):
    """
    This functions computes the diarization error rate (DER) relying on the dscore package: https://github.com/nryant/dscore
    Lots of code are copied from the referred Github.
    This package removes the RTTM dependency and compute the DER directly on the python list, which is more user-friendly.

    Args:
        hyp_turns (list): a list of tuple of (onset, offset, speaker_id, recording_id)
        ref_turns (list): same format as hyp_turns
        step (float): step size in seconds for scoring
        jer_min_ref_dur (float): minimum reference speaker duration for JER
        collar (float): collar size in seconds for DER computaton
        ignore_overlaps (bool): ignore overlaps when computing DER

    Return:
        der score (float)

    Note:
        The units of the onset and offset are second.
    """

    def type_convert(turns):
        return [
            Turn(onset=on, offset=off, speaker_id=spk, file_id=reco)
            for on, off, spk, reco in turns
        ]

    sys_turns = type_convert(hyp_turns)
    ref_turns = type_convert(ref_turns)

    warn(
        "No universal evaluation map specified. Approximating from "
        "reference and speaker turn extents..."
    )
    uem = gen_uem(ref_turns, sys_turns)

    # Trim turns to UEM scoring regions and merge any that overlap.
    info("Trimming reference speaker turns to UEM scoring regions...", file=sys.stderr)
    ref_turns = trim_turns(ref_turns, uem)
    info("Trimming system speaker turns to UEM scoring regions...", file=sys.stderr)
    sys_turns = trim_turns(sys_turns, uem)
    info("Checking for overlapping reference speaker turns...", file=sys.stderr)
    ref_turns = merge_turns(ref_turns)
    info("Checking for overlapping system speaker turns...", file=sys.stderr)
    sys_turns = merge_turns(sys_turns)

    # Score.
    check_for_empty_files(ref_turns, sys_turns, uem)
    file_scores, global_scores = score(
        ref_turns,
        sys_turns,
        uem,
        step=step,
        jer_min_ref_dur=jer_min_ref_dur,
        collar=collar,
        ignore_overlaps=ignore_overlaps,
    )
    return global_scores.der


def der_for_dynamic_superb(
    hyps: List[str],
    refs: List[str],
):
    """
    This function computes the diarization error rate (DER) relying on the `der` function in this package.
    The difference lies in the input format. This function takes the input in pure text to be compatible with Dynamic-SUPERB.

    Each hyp and ref should be a list of speaking turns in the format of
        '(spk1, onset1, offset1) (spk2, onset2, offset2) (spk3, onset3, offset3) ...'

    Usually, the spk labels of some turns are the same, and the unit of the onset and offset is second. For example:
        '(spk1, 0.02, 3.23) (spk2, 4.12, 4.98) (spk1, 5.32, 6.72) (spk2, 5.52, 8.24)'

    Args:
        preds (List[str]): the predicted segments of each recording in text
        refs (List[str]): the reference segments of each recording in text

    Return:
        der score (float)
    """

    def parse_text_into_turns(text_turn: str):
        text_turn = re.sub(r" +", r" ", text_turn)
        turns = re.findall(r"\(\S+,\s?\S+,\s?\S+\)", text_turn)

        spks = []
        starts = []
        ends = []
        for turn in turns:
            turn = re.sub(r" +", r" ", turn)
            spk, start, end = re.search(r"\((\S+),\s?(\S+),\s?(\S+)\)", turn).groups()
            spks.append(spk)
            starts.append(float(start))
            ends.append(float(end))

        return list(zip(spks, starts, ends))

    def parse_turns_with_fileid(all_text_turns: List[str]):
        all_turns = []
        for fileid, text_turn in enumerate(all_text_turns):
            for spk, start, end in parse_text_into_turns(text_turn):
                all_turns.append((start, end, f"{fileid}-{spk}", str(fileid)))
        return all_turns

    hyp_turns = parse_turns_with_fileid(hyps)
    ref_turns = parse_turns_with_fileid(refs)

    return der(hyp_turns, ref_turns)



from typing import List, Union
import editdistance as ed


def ter(hyps: List[Union[str, List[str]]], refs: List[Union[str, List[str]]]) -> float:
    """Token error rate calculator.

    Args:
        hyps (List[Union[str, List[str]]]): List of hypotheses.
        refs (List[Union[str, List[str]]]): List of references.

    Returns:
        float: Averaged token error rate overall utterances.
    """
    error_tokens = 0
    total_tokens = 0
    for h, r in zip(hyps, refs):
        error_tokens += ed.eval(h, r)
        total_tokens += len(r)
    return float(error_tokens) / float(total_tokens)


def wer(hyps: List[str], refs: List[str]) -> float:
    """Word error rate calculator.

    Args:
        hyps (List[str]): List of hypotheses.
        refs (List[str]): List of references.

    Returns:
        float: Averaged word error rate overall utterances.
    """
    hyps = [h.upper().split(" ") for h in hyps]
    refs = [r.upper().split(" ") for r in refs]
    return ter(hyps, refs)


def per(hyps: List[str], refs: List[str]) -> float:
    """
    This function implements the phone error rate (PER).
    The phonetic transcription should be denoted in ARPAbet symbols and space-separated.
    For example:
        "AH0 K OW1 L D L UW1 S AH0 D IH0 N D IH1 F ER0 AH0 N S R EY1 N D IH0 N HH IH0 Z S OW1 L"

    Args:
        preds: A list of predicted phonetic transcription
        refs: A list of reference phonetic transcription
    
    Return: The corpus-level phone error rate (PER)
    """
    return wer(hyps, refs)


def cer(hyps: List[str], refs: List[str]) -> float:
    """Character error rate calculator.

    Args:
        hyps (List[str]): List of hypotheses.
        refs (List[str]): List of references.

    Returns:
        float: Averaged character error rate overall utterances.
    """
    hyps = [h.upper() for h in hyps]
    refs = [r.upper() for r in refs]
    return ter(hyps, refs)



from speech_metrics import wer, cer


def test_wer():
    ref = ["Hello how are you today"]
    hyp1 = ["HELLO how are you today"]
    hyp2 = ["HELLO how are you Today"]
    assert wer(hyp1, ref) == wer(hyp2, ref)


def test_cer():
    ref = ["Hello how are you today"]
    hyp1 = ["HELLO how are you today"]
    hyp2 = ["HELLO how are you Today"]
    assert cer(hyp1, ref) == cer(hyp2, ref)



"""
Original Author: Yung-Sung Chuang
Modified Author: Shu-wen Yang
"""

from typing import List
from collections import defaultdict

from .common import cer


def parse(format: str):
    output = defaultdict(list)
    for single in format.split(", "):
        slot_type, slot_value = single.strip().split(": ")
        output[slot_type.strip()].append(slot_value.strip())
    return output


def slot_type_f1(preds: List[str], refs: List[str]) -> float:
    """
    This function implements the slot type F1 metric for the slot filling task.
    A single pred or a ref should be in the format of "<slot type 1>: <slot value 1>, <slot type 2>: <slot value 2>, ..."

    Args:
        preds: A list of predicted slot types and values
        refs: A list of reference slot types and values

    Return: The F1 score
    """
    F1s = []
    for p, t in zip(preds, refs):
        hyp_dict = parse(p)
        ref_dict = parse(t)

        # Slot Type F1 evaluation
        if len(hyp_dict.keys()) == 0 and len(ref_dict.keys()) == 0:
            F1 = 1.0
        elif len(hyp_dict.keys()) == 0:
            F1 = 0.0
        elif len(ref_dict.keys()) == 0:
            F1 = 0.0
        else:
            P, R = 0.0, 0.0
            for slot in ref_dict:
                if slot in hyp_dict:
                    R += 1
            R = R / len(ref_dict.keys())
            for slot in hyp_dict:
                if slot in ref_dict:
                    P += 1
            P = P / len(hyp_dict.keys())
            F1 = 2 * P * R / (P + R) if (P + R) > 0 else 0.0
        F1s.append(F1)

    return sum(F1s) / len(F1s)


def slot_value_cer(hypothesis: List[str], groundtruth: List[str], **kwargs) -> float:
    """
    This function implements the slot value CER metric for the slot filling task.
    A single pred or a ref should be in the format of "<slot type 1>: <slot value 1>, <slot type 2>: <slot value 2>, ..."

    Args:
        preds: A list of predicted slot types and values
        refs: A list of reference slot types and values

    Return: The CER score
    """
    value_hyps, value_refs = [], []
    for p, t in zip(hypothesis, groundtruth):
        ref_dict = parse(p)
        hyp_dict = parse(t)

        # Slot Value WER/CER evaluation
        unique_slots = list(ref_dict.keys())
        for slot in unique_slots:
            for ref_i, ref_v in enumerate(ref_dict[slot]):
                if slot not in hyp_dict:
                    hyp_v = ""
                    value_refs.append(ref_v)
                    value_hyps.append(hyp_v)
                else:
                    min_cer = 100
                    best_hyp_v = ""
                    for hyp_v in hyp_dict[slot]:
                        tmp_cer = cer([hyp_v], [ref_v])
                        if min_cer > tmp_cer:
                            min_cer = tmp_cer
                            best_hyp_v = hyp_v
                    value_refs.append(ref_v)
                    value_hyps.append(best_hyp_v)

    return cer(value_hyps, value_refs)



"""Utility functions."""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import itertools
import sys

from . import six

__all__ = ['clip', 'error', 'format_float', 'groupby', 'info', 'warn', 'xor']


def error(msg, file=sys.stderr):
    """Log error message ``msg`` to stderr."""
    msg = 'ERROR: %s' % msg
    if six.PY2:
        msg = msg.encode('utf-8')
    print(msg, file=file)


def info(msg, print_level=False, file=sys.stdout):
    """Log info message ``msg`` to stdout."""
    if print_level:
        msg = 'INFO: %s' %msg
    if six.PY2:
        msg = msg.encode('utf-8')
    print(msg, file=file)


def warn(msg, file=sys.stderr):
    """Log warning message ``msg`` to stderr."""
    msg = 'WARNING: %s' %msg
    if six.PY2:
        msg = msg.encode('utf-8')
    print(msg, file=file)


def xor(x, y):
    """Return truth value of ``x`` XOR ``y``."""
    return bool(x) != bool(y)


def format_float(x, n_digits=3):
    """Format floating point number for output as string.

    Parameters
    ----------
    x : float
        Number.

    n_digits : int, optional
        Number of decimal digits to round to.
        (Default: 3)

    Returns
    -------
    s : str
        Formatted string.
    """
    fmt_str = '%%.%df' % n_digits
    return fmt_str % round(x, n_digits)


def clip(x, lower, upper):
    """Clip ``x`` to [``lower``, ``upper``]."""
    return min(max(x, lower), upper)


def groupby(iterable, keyfunc):
    """Wrapper around ``itertools.groupby`` which sorts data first."""
    iterable = sorted(iterable, key=keyfunc)
    for key, group in itertools.groupby(iterable, keyfunc):
        yield key, group



import re
import argparse
from pathlib import Path
from collections import defaultdict

from speech_metrics import der, der_for_dynamic_superb
from speech_metrics.rttm import rttm_to_turns, turns_to_text


def test_ders():
    ref = "data/der/ref.rttm"
    hyp = "data/der/hyp-der-7.53.rttm"

    ref_turns = rttm_to_turns(ref)
    hyp_turns = rttm_to_turns(hyp)

    ref_texts = turns_to_text(ref_turns)
    hyp_texts = turns_to_text(hyp_turns)

    recos = sorted(ref_texts.keys())
    ref_texts = [ref_texts[reco] for reco in recos]
    hyp_texts = [hyp_texts[reco] for reco in recos]

    assert round(der(hyp_turns, ref_turns), 2) == 7.53
    assert round(der_for_dynamic_superb(hyp_texts, ref_texts), 2) == 7.53



import numpy as np

from speech_metrics import pesq, stoi, sisdri

SAMPLE_RATE = 16000


def test_pesq():
    wav_hyp = np.random.randn(SAMPLE_RATE * 12)
    wav_ref = np.random.randn(SAMPLE_RATE * 12)
    assert isinstance(pesq([wav_hyp] * 2, [wav_ref] * 2), float)


def test_stoi():
    wav_hyp = np.random.randn(SAMPLE_RATE * 12)
    wav_ref = np.random.randn(SAMPLE_RATE * 12)
    assert isinstance(stoi([wav_hyp] * 2, [wav_ref] * 2), float)


def test_sisdri():
    wav_mix = np.random.randn(SAMPLE_RATE * 12)
    wav_hyp = np.random.randn(2, SAMPLE_RATE * 12)
    wav_ref = np.random.randn(2, SAMPLE_RATE * 12)
    assert isinstance(sisdri([wav_mix] * 2, [wav_hyp] * 2, [wav_ref] * 2), float)



"""Custom argument parser and action classes."""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
import argparse
import sys


__all__ = ['ArgumentParser']


class ArgumentParser(argparse.ArgumentParser):
    """Sub-class of ``ArgumentParser`` that write errors to STDERR."""
    def error(self, message):
        sys.stderr.write('error: %s\n' % message)
        self.print_help()
        sys.exit(2)



"""
This is a direct copy of the metric.py file in the Asteroid toolkit.
The copy is beneficial to disentangle the metric function in Asteroid from the other many functions.

Original author: The Asteroid Team
Modified author: Shu-wen Yang 
"""

import warnings
import traceback

import numpy as np
from pb_bss_eval import InputMetrics, OutputMetrics

ALL_METRICS = ["si_sdr", "sdr", "sir", "sar", "stoi", "pesq"]


def average_arrays_in_dic(dic):
    """Take average of numpy arrays in a dictionary.

    Args:
        dic (dict): Input dictionary to take average from

    Returns:
        dict: New dictionary with array averaged.

    """
    # Copy dic first
    dic = dict(dic)
    for k, v in dic.items():
        if isinstance(v, np.ndarray):
            dic[k] = float(v.mean())
    return dic


def get_metrics(
    mix,
    clean,
    estimate,
    sample_rate=16000,
    metrics_list="all",
    average=True,
    compute_permutation=False,
    ignore_metrics_errors=False,
    filename=None,
):
    r"""Get speech separation/enhancement metrics from mix/clean/estimate.

    Args:
        mix (np.array): mixture array.
        clean (np.array): reference array.
        estimate (np.array): estimate array.
        sample_rate (int): sampling rate of the audio clips.
        metrics_list (Union[List[str], str): List of metrics to compute.
            Defaults to 'all' (['si_sdr', 'sdr', 'sir', 'sar', 'stoi', 'pesq']).
        average (bool): Return dict([float]) if True, else dict([array]).
        compute_permutation (bool): Whether to compute the permutation on
            estimate sources for the output metrics (default False)
        ignore_metrics_errors (bool): Whether to ignore errors that occur in
            computing the metrics. A warning will be printed instead.
        filename (str, optional): If computing a metric fails, print this
            filename along with the exception/warning message for debugging purposes.

    Shape:
        - mix: :math:`(D, N)` or `(N, )`.
        - clean: :math:`(K\_source, N)` or `(N, )`.
        - estimate: :math:`(K\_target, N)` or `(N, )`.

    Returns:
        dict: Dictionary with all requested metrics, with `'input_'` prefix
        for metrics at the input (mixture against clean), no prefix at the
        output (estimate against clean). Output format depends on average.

    Examples
        >>> import numpy as np
        >>> import pprint
        >>> from asteroid.metrics import get_metrics
        >>> mix = np.random.randn(1, 16000)
        >>> clean = np.random.randn(2, 16000)
        >>> est = np.random.randn(2, 16000)
        >>> metrics_dict = get_metrics(mix, clean, est, sample_rate=8000,
        ...                            metrics_list='all')
        >>> pprint.pprint(metrics_dict)
        {'input_pesq': 1.924380898475647,
         'input_sar': -11.67667585294225,
         'input_sdr': -14.88667106190552,
         'input_si_sdr': -52.43849784881705,
         'input_sir': -0.10419427290163795,
         'input_stoi': 0.015112115177091223,
         'pesq': 1.7713886499404907,
         'sar': -11.610963379923195,
         'sdr': -14.527246041125844,
         'si_sdr': -46.26557128489802,
         'sir': 0.4799929272243427,
         'stoi': 0.022023073540350643}

    """
    if metrics_list == "all":
        metrics_list = ALL_METRICS
    if isinstance(metrics_list, str):
        metrics_list = [metrics_list]
    # For each utterance, we get a dictionary with the input and output metrics
    input_metrics = InputMetrics(
        observation=mix,
        speech_source=clean,
        enable_si_sdr=True,
        sample_rate=sample_rate,
    )
    output_metrics = OutputMetrics(
        speech_prediction=estimate,
        speech_source=clean,
        enable_si_sdr=True,
        sample_rate=sample_rate,
        compute_permutation=compute_permutation,
    )
    utt_metrics = {}
    for src, prefix in [(input_metrics, "input_"), (output_metrics, "")]:
        for metric in metrics_list:
            # key: eg. "input_pesq" or "pesq"
            key = prefix + metric
            try:
                utt_metrics[key] = src[metric]
            except Exception as err:
                if ignore_metrics_errors:
                    warnings.warn(
                        f"Error computing {key} for {filename or '<unknown file>'},"
                        f" ignoring. Error was: {err}",
                        RuntimeWarning,
                    )
                    traceback.print_stack()
                    utt_metrics[key] = None
                else:
                    raise RuntimeError(
                        f"Error computing {key} for {filename or '<unknown file>'}"
                    ) from err
    if average:
        return average_arrays_in_dic(utt_metrics)
    else:
        return utt_metrics



from speech_metrics import sacre_bleu


def test_bleu():
    _refs = [
        "The dog bit the man.",
        "It was not unexpected.",
        "The man bit him first.",
    ]
    _hyps = [
        "The dog bit the man.",
        "It wasn't surprising.",
        "The man had just bitten him.",
    ]
    assert isinstance(sacre_bleu(_hyps, _refs), float)



"""
Author: Shu-wen Yang
"""


import re
from typing import List, Tuple
from collections import defaultdict


def rttm_to_turns(rttm_path: str):
    """
    Args:
        rttm_path (str): the path of the RTTM file

    Return: (List[Tuple[float, float, str, str]])
        The segment information from the RTTM file.
        Each tuple is (start_sec, end_sec, speaker_id, recording_id)
    """
    turns = []

    with open(rttm_path, "r") as f:
        for line in f.readlines():
            line = line.strip()
            line = re.sub(r" +", " ", line)
            line = re.sub(r"\t", " ", line)
            _, reco_id, _, on, dur, _, _, spk, _ = line.split(" ")

            start = float(on)
            end = float(on) + float(dur)

            turns.append((start, end, spk, reco_id))

    return turns


def turns_to_text(turns: List[Tuple[float, float, str, str]]):
    """
    Organize the turns based on the recording id.
    Each recording gets a string describing the turns in the form of:
        '(speaker_id 1, start_sec 1, end_sec 1) (speaker_id 2, start_sec 2, end_sec 2) ...'

    Args:
        turns (List[Tuple[float, float, str, str]]): The output of `rttm_to_turns`

    Return: Dict[str, str]
        map the recording id to the string describing the turns
    """
    reco2turns = defaultdict(list)
    for start, end, spk, reco in turns:
        reco2turns[reco].append((spk, start, end))

    reco2text = {}
    for reco, segs in reco2turns.items():
        segs = [
            f"({spk}, {round(start, 2)}, {round(end, 2)})" for spk, start, end in segs
        ]
        text = " ".join(segs)
        reco2text[reco] = text

    return reco2text



from pathlib import Path
from setuptools import setup, find_packages

here = Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name='speech_metrics',
    version='0.2.2',
    url='https://github.com/leo19941227/speech_metrics',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Shu-wen Yang',
    author_email='leo19941227@gmail.com',
    description='A lightweight package for common metrics in speech processing',
    license='MIT License',
    packages=find_packages(),
    install_requires=[
        "editdistance",
        "pb_bss_eval",
        "sacrebleu",
    ],
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
